"Python3 document categorization library, keytoks method."

##  DOCALIB :
##
##  Python3 data & routines to supply to toccata test harness s/w :
##  Copyright 2015, R.S. Forsyth.
##  Released under GNU Public licence, version 3:
##  http://www.gnu.org/licenses/gpl.txt
##  first version : 12/12/2013
##  last revision : 24/04/2017
##  tokens selected by discriminability, rankcor scoring.


ShowGNU = 0  ##  change to 1 2c licence statement
GNUblurb = """
  This copyrighted program is free software: you can redistribute it &/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
"""
if ShowGNU:
    print(GNUblurb)

import  math, os, string, sys

if sys.platform.startswith("linu"):  ##  presume linux
    sys.path.append("~/toccata/p3")
else:  ##  presume windows
    sys.path.append("c:\\toccata\\p3")
    

from globdefs import *
from glib import  add2dic, numerize
from tocclib import  bastats
from tocclib import  corsums, corrval
from tocclib import  ranx
from tocclib import  kagdist, midvalue


##  global vars :
Showoff = 0  ##  >0 for debugging o/p
Gold =  5.0 ** 0.5 * 0.5 + 0.5  ##  good approximation to golden number


##  functions first, then class(es) :

##  GETFREX :
def getfrex (toklist):
    "makes freqency table for a single token list."

    freqtab = {}
    for t in toklist:
        f = freqtab.get(t,0)
        freqtab[t] = f + 1

    return  freqtab


##  TOPTOKS :
def toptoks (freqtab,h=144):
    "returns set of high-frequency tokens from frequency dict."

    v = len(freqtab)  ##  vocsize
    sortlist = []
    for k in freqtab:
        f = freqtab[k]
        if f > 1:  ## always skip hapaxes
            sortlist.append((f,k))
    sortlist.sort(reverse=True)

    return  sortlist[:h]


##  ADDSNIPS :
def addsnips (d,snipsize):
    "adds snippet data to text d if not already present."

    d.snippets = 0
    d.snipsets = []  ##  list of sets
    minsnip = int(round(snipsize / Gold))
    for p in range(0,d.size,snipsize):
        slab = d.toklist[p:p+snipsize]
        if len(slab) >= minsnip or p==0:
            d.snipsets.append(set(slab))
            d.snippets += 1

    return  ##  side-effect on d.


####  DOCADAT :
class Docadat (object):
    ##  collection of data+methods to do document classification :
    ##  assumes access to 2 Sack() objects as args, doclist & paradat.
    ##  doclist attributes expected include:
    ##    classnum, freqtab, outcome, size, text, textlen, toklist,
    ##  paradat attributes expected include:
    ##    catlist, cats, paraline, testmeta, trainmet,

    ##  N.B. this library only to make changes in caller if clearly innocuous!
    

    def __init__ (self,paradat,doclist):
        "initialization method for Docadat objects."

        self.classes = paradat.cats
        self.catlist = paradat.catlist
        self.id = paradat.trainmet+paradat.testmeta
        self.alldocs = len(doclist)  ##  number of training-corpus documents

        return

    def loadpars (self,paradat,sep1=' ',sep2='='):
        "loads extra docalib parameters if any, provided in paraline."

        parline = paradat.paraline.lower().strip()
        self.pars = Sack(parline)
        parbits = parline.split(sep1)  ##  item separator defaults to space
        for pb in parbits:
            bits = pb.split(sep2)  ##  internal separator (e.g. "gramsize=4")
            if len(bits) == 2:
                v = numerize(bits[1])
                if type(v) == type(9) or type(v) == type(0.5):  ##  numeric
                    setattr(self.pars,bits[0],v)
                else:  ##  nonnumeric
                    setattr(self.pars,bits[0],bits[1])
                ##  no limits imposed !
        ##  deal with internal parameters :
        ss = 115  ##  default snipsize
        if hasattr(self.pars,"snipsize"):
            self.pars.snipsize = numerize(getattr(self.pars,"snipsize",ss))
        else:
            self.pars.snipsize = ss  ##  may still be altered if unrealistic
        if hasattr(self.pars,"topkeys"):
            self.pars.topkeys = numerize(getattr(self.pars,"topkeys",0))
        else:
            self.pars.topkeys = 0
        ##  will adjust silly values (outside 8..256) after snippet processing.
        if hasattr(self.pars,"cormode") and self.pars.cormode == "pm":
            self.pars.rankcor = 0  ##  product-moment correlation
        else:
            self.pars.rankcor = 1  ##  rankcor
            
        return  ##  doesn't crash if paradat.paraline empty

    def modprep (self,paradat):
        "does once-only overall tasks."

        ##  check that snipsize is reasonable :
        ns = paradat.tottoks / self.pars.snipsize
        if ns < 64 or ns > 8192:
            self.pars.snipsize = int(paradat.tottoks ** 0.5)
            print("snipsize altered to",self.pars.snipsize)
        else:
            print("snipsize =",self.pars.snipsize)
        ##  attach snippets to docs :
        totsnips = 0
        minsnip = int(round(self.pars.snipsize / Gold))
        for d in paradat.doclist:  ##  deals with entire sample
            d.snippets = 0
            d.snipsets = []  ##  list of sets
            for p in range(0,d.size,self.pars.snipsize):
                slab = d.toklist[p:p+self.pars.snipsize]
                if len(slab) >= minsnip or p==0:
                    d.snipsets.append(set(slab))
                    d.snippets += 1
                    totsnips += 1
        print("Number of snippets =",totsnips)
        paradat.totsnips = totsnips  ##  does belong to whole corpus
        ##  now find high-freq tokens :
        tokfrex = {}
        for d in paradat.doclist:
            for snip in d.snipsets:
                for t in snip:
                    if t in tokfrex:
                        tokfrex[t] += 1
                    else:
                        tokfrex[t] = 1
            d.snippets = len(d.snipsets)  ##  record number of snippets in d
        paradat.highfrex = toptoks(tokfrex,1024)
                
        return

    def makemods (self,paradat,doclist):
        "builds classification models for each category."

        self.modlist = []  ##  list of category models
        self.modeltab = {}  ##  dictionary of models, if lookup by name needed
        self.name2num = {}
        self.vocsize = 0 ; self.tottoks = 0
        for d in doclist:
            self.tottoks += d.size
            self.vocsize += len(d.freqtab)
        ##  establish suitable topkeys :
        if self.pars.topkeys < 8:
            self.pars.topkeys = max(8,int(math.sqrt(self.vocsize)))
        if self.pars.topkeys > 256:
            self.pars.topkeys = 256
        ##  self.pars.topkeys = int(math.sqrt(self.vocsize))
        self.quayset = set()
        cn = 0
        for c in self.catlist:  ##  relies on consistent order of catlist
            m = self.makemod(c,cn,paradat,doclist)
            self.modlist.append(m)
            self.modeltab[c] = m  ##  also point from dict keyed by catname
            self.name2num[c] = cn
            cn += 1  ##  keep track of position in model list
            m.keyfrex = {}  ##  to save recomputing
            for bk in m.bestkeys:
                k = bk[-1]  ##  last item in tuple
                self.quayset.add(k)
                m.keyfrex[k] = bk[2]  ##  third item must be in-class freq !!
        if Showoff > 1:
            print(self.quayset)
        if Showoff > 0:
            print(len(self.quayset))
        ##  make sure all models have correct snippet frex :
        for cn,c in enumerate(paradat.catlist):
            m = self.modlist[cn]
            for q in self.quayset:
                m.keyfrex[q] = self.snipsum(q,c,doclist)
        
        self.mods = len(self.modlist)
        assert  self.mods == self.classes
        self.docs = len(doclist)
        
        return

    def snipsum (self,q,c,doclist):
        "counts snippet frequency of token q in class c."

        f = 0
        for d in doclist:
            if d.outcome == c:
                for s in d.snipsets:
                    if q in s:
                        f += 1

        return  f

    def makemod (self,c,cn,paradat,doclist):
        "builds model info for a single category."

        m = Sack(c)  ##  gathered into a Sack() object
        m.tottoks = 0
        m.catname = c
        m.classnum = cn
        m.totsnips = 0
        for d in doclist:
            if d.outcome == c:
                m.tottoks += d.size
                m.totsnips += d.snippets
        ##  pick best keytoks :
        bestlist = self.keytoks(c,cn,paradat,doclist)
        tk = self.pars.topkeys  ##  / 2
        ##  poskeys = tk / Gold
        ##  negkeys = max(1.0,tk - poskeys)
        poskeys = tk ; negkeys = max(1,tk / Gold)
        m.bestkeys = bestlist[:int(poskeys)] + bestlist[-int(negkeys):]
        if Showoff > 1:  ##  testing :
            print(m.classnum,m.catname,self.pars.topkeys)
            for k in m.bestkeys:
                print(k)
                                   
        return  m

    def keytoks (self,c,cn,paradat,doclist):
        "returns freq toks sorted by pq-quayness in class c."

        sortlist = []
        for sf,t in paradat.highfrex:  ##  only look at relatively common tokens
            f1 = 0 ; f2 = 0
            n1 = 0 ; n2 = 0
            for d in doclist:
                f = 0
                for s in d.snipsets:
                    if t in s:
                        f += 1
                if d.outcome == c:
                    f1 += f
                    n1 += d.snippets
                else:
                    f2 += f
                    n2 += d.snippets
            if f1+f2 > 0:  ##  zero freq possible in subset
                r1 = f1 / max(1,n1)
                r2 = f2 / max(1,n2)
                quayval = r1 * (1.0-r2)
                sortlist.append((100.0*quayval,r1-r2,f1,f2,sf,t))

        return  sorted(sortlist,reverse=True)

    def showmod (self,cn,fo=sys.stdout):
        "shows info related to model of category number cn."

        m = self.modlist[cn]
        assert cn == m.classnum
        print("category:",m.catname,'(',cn,')',file=fo)
        print("snippets =",m.totsnips,file=fo)
        print("tokens =",m.tottoks,NL,file=fo)
        for v in sorted(self.quayset):
            quayval = m.keyfrex.get(v,0)
            sr = quayval / max(1,m.totsnips)
            lout = "{0:15s} {1:7d} {2:8.4f}".format(v,quayval,sr*100.0)
            print(lout,file=fo)
        print('',file=fo)

        return

    def showmods (self,paradat,fo=sys.stdout):
        "shows models of all categories."

        print("classes =",self.classes,file=fo)
        print(paradat.catlist,file=fo)
        print("docs =",paradat.docs,file=fo)
        print("vocsize =",self.vocsize,file=fo)
        print("tokens =",paradat.tottoks,file=fo)
        print("keys used =",len(self.quayset),file=fo)
        print(NL,file=fo)
        for cn in range(self.classes):
            self.showmod(cn,fo)

        return
        ##  doesn't open or close o/p file; caller must do that.

    def corrsim (self,cn,thisdoc,quayset,paradat):
        "matches category model to given document."
        ##  uses correlation of quay snipfrex.

        nv = len(self.quayset)
        if Showoff > 0:
            print(cn,thisdoc.name,nv,file=paradat.logfile)
        xvec = [0] * nv ; yvec = [0] * nv
        j = 0 ; m = self.modlist[cn]
        for quay in self.quayset:
            xvec[j] = m.keyfrex.get(quay,0)  ##  snipfreq in class
            for s in thisdoc.snipsets:  ##  not currently saved, so recount
                if quay in s:
                    yvec[j] += 1
            ##  too many nils with small docs ??
            if Showoff > 0 and j < 26:
                print(j,xvec[j],yvec[j],file=paradat.logfile)  ##  checking
            j += 1
        if self.pars.rankcor > 0:
            cx = corsums(ranx(xvec),ranx(yvec))
        else:
            cx = corsums(xvec,yvec)
        
        return  corrval(cx)  ##  matching score, higher implies stronger match
          
    def modsims (self,thisdoc,paradat):
        "returns association score of each model to this doc."

        if not hasattr(thisdoc,"snipsets"):  ##  cd be absent from holdout docs
            addsnips(thisdoc,self.pars.snipsize)
        catvals = [0.0] * paradat.cats
        for cpos,c in enumerate(paradat.catlist):
            ##  m = self.modlist[cpos]
            catvals[cpos] = self.corrsim(cpos,thisdoc,self.quayset,paradat)

        if sum(catvals) <= 0.0:
            print("!!!!",thisdoc.name,thisdoc.snippets,catvals)
        return  catvals  ##  no scaling to quasi-probs at present.
        ##  also no inversion if distances used, so more must mean better.

    
